@@ -1,4 +1,4 @@
-# Copyright (c) 2011 CentralNic Ltd. All rights reserved. This program is
+# Copyright (c) 2012 CentralNic Ltd. All rights reserved. This program is
 # free software; you can redistribute it and/or modify it under the same
 # terms as Perl itself.
 # 
@@ -184,9 +184,11 @@ certificate information). See the relevant manpage for examples.
 This method will C<croak()> if connection fails, so be sure to use C<eval()> if
 you want to catch the error.
 
-The return value for C<connect()> will be the EPP E<lt>greetingE<gt> frame
-returned by the server. Please note that the same caveat about blocking applies
-to this method as to C<get_frame()> (see below).
+By default, the return value for C<connect()> will be the EPP E<lt>greetingE<gt>
+frame returned by the server. Please note that the same caveat about blocking
+applies to this method as to C<get_frame()> (see below).
+
+If you want to get the greeting yourself, set C<$params{no_greeting}>.
 
 =cut
 
@@ -201,7 +203,7 @@ sub connect {
 
 	}
 
-	return $self->get_frame;
+	return ($params{'no_greeting'} ? 1 : $self->get_frame);
 
 }
 
@@ -409,7 +411,7 @@ CentralNic Ltd (L<http://www.centralnic.com/>).
 
 =head1 COPYRIGHT
 
-This module is (c) 2011 CentralNic Ltd. This module is free software; you can
+This module is (c) 2012 CentralNic Ltd. This module is free software; you can
 redistribute it and/or modify it under the same terms as Perl itself.
 
 =head1 SEE ALSO
